@@ -3,6 +3,7 @@
 #include <config.h>
 #endif
 
+#include <time.h>
 #include <gio/gio.h>
 #include <ibus.h>
 #include <m17n.h>
@@ -443,6 +444,8 @@ ibus_m17n_engine_init (IBusM17NEngine *m17n)
     g_object_ref_sink (m17n->table);
     m17n->context = NULL;
     m17n->us_keymap = ibus_keymap_get ("us");
+    /* Load $HOME/.XCompose file: */
+    ibus_engine_simple_add_table_by_locale ((IBusEngineSimple *) m17n, NULL);
 }
 
 static GObject*
@@ -550,6 +553,10 @@ ibus_m17n_engine_update_preedit (IBusM17NEngine *m17n)
     gchar *buf;
     IBusM17NEngineClass *klass = (IBusM17NEngineClass *) G_OBJECT_GET_CLASS (m17n);
 
+    if (!mtext_len (m17n->context->preedit)) {
+        /* Do not update the preedit if it has length 0 to avoid flicker */
+        return;
+    }
     buf = ibus_m17n_mtext_to_utf8 (m17n->context->preedit);
     if (buf) {
         text = ibus_text_new_from_string (buf);
@@ -564,12 +571,27 @@ ibus_m17n_engine_update_preedit (IBusM17NEngine *m17n)
         ibus_engine_update_preedit_text_with_mode ((IBusEngine *) m17n,
                                                    text,
                                                    m17n->context->cursor_pos,
-                                                   mtext_len (m17n->context->preedit) > 0,
+                                                   TRUE,
                                                    klass->preedit_focus_mode);
     }
     g_free (buf);
 }
 
+static void
+ibus_m17n_engine_hide_preedit_if_empty (IBusM17NEngine *m17n)
+{
+    IBusM17NEngineClass *klass = (IBusM17NEngineClass *) G_OBJECT_GET_CLASS (m17n);
+    if (mtext_len (m17n->context->preedit)) {
+        return;
+    }
+    ibus_engine_update_preedit_text_with_mode (
+        (IBusEngine *) m17n,
+        ibus_text_new_from_string (""),
+        0,
+        FALSE,
+        klass->preedit_focus_mode);
+}
+
 static void
 ibus_m17n_engine_commit_string (IBusM17NEngine *m17n,
                                 const gchar    *string)
@@ -699,10 +721,12 @@ ibus_m17n_engine_process_key (IBusM17NEngine *m17n,
     gchar *buf;
     MText *produced;
     gint retval;
+    gchar *sym_name = msymbol_name (key);
 
     retval = minput_filter (m17n->context, key, NULL);
 
     if (retval) {
+        ibus_m17n_engine_hide_preedit_if_empty (m17n);
         return TRUE;
     }
 
@@ -717,9 +741,60 @@ ibus_m17n_engine_process_key (IBusM17NEngine *m17n,
     buf = ibus_m17n_mtext_to_utf8 (produced);
     m17n_object_unref (produced);
 
+    if (retval && buf && strlen (buf)) {
+        /*
+          Prefer commit to "return FALSE;" for space and other
+          keys where the msymbol name is exactly one character to
+          avoid ordering problems in Mutter, see:
+          https://github.com/ibus/ibus-m17n/issues/72
+
+          The keys where the msymbol name is exactly one character
+          should include all keys which result in just a plain Unicode
+          value with no modifiers pressed (Key combinations with
+          modifiers have msymbol names longer than one character, for
+          example Control+a has the msymbol name "C-a") and it should
+          exclude all control characters like Return and Tab.
+        */
+        gchar *commit_string = NULL;
+        if (strlen (sym_name) == 1) {
+            commit_string = g_strconcat (buf, sym_name, NULL);
+        }
+        else if (g_strcmp0 (sym_name, "KP_Space") == 0) {
+            commit_string = g_strconcat (buf, " ", NULL);
+        }
+        if (commit_string) {
+            ibus_m17n_engine_commit_string (m17n, commit_string);
+            g_free (commit_string);
+            g_free (buf);
+            ibus_m17n_engine_hide_preedit_if_empty (m17n);
+            return TRUE;
+        }
+    }
+
     if (buf && strlen (buf)) {
         ibus_m17n_engine_commit_string (m17n, buf);
     }
+
+    ibus_m17n_engine_hide_preedit_if_empty (m17n);
+
+    if (retval && buf && strlen (buf)) {
+        /*
+          We have a key event here which caused a commit
+          but handling it by appending to the commit string
+          was not possible (For example Return or KP_Enter
+          or keys where modifiers were pressed).
+
+          But a sleep of 0.1 seconds between the commit and passing
+          through that key event helps Mutter to guess which actions
+          belong together and get the order right. So this mostly
+          "fixes" the problem as well.  A sleep is not so nice, but as
+          these key events are rare that sleep should not hurt much.
+        */
+        struct timespec delay;
+        delay.tv_sec = 0;
+        delay.tv_nsec = 100000000; // 100,000,000 nanoseconds = 0.1 seconds
+        nanosleep (&delay, NULL);
+    }
     g_free (buf);
 
     return retval == 0;
@@ -747,28 +822,28 @@ ibus_m17n_engine_process_key_event (IBusEngine     *engine,
         break;
     }
 
+    if (klass->use_us_layout) {
+        if (g_strcmp0 (ibus_keyval_name (keyval), "Multi_key") != 0) {
+            /*
+              Do not translate the Multi_key: If the non-US layout has
+              a Multi_key, trying to translate it to US layout just
+              takes it away. Skipping the translation keeps the
+              Multi_key around which is more useful, it can still be
+              used for Compose then.
+             */
+            keyval = ibus_keymap_lookup_keysym (m17n->us_keymap,
+                                                keycode,
+                                                modifiers);
+        }
+    }
+
     /*
       Compose support:
 
       IBusM17NEngine inherits from
       IBusEngineSimple. IBUS_ENGINE_CLASS(parent_class)->process_key_event()
-      calls ibus_engine_simple_process_key_event(). If this does not
-      start to process a compose sequence, it calls
-      ibus_engine_hide_preedit_text((IBusEngine *)simple) before
-      returning FALSE. If m17n->context->preedit was not empty when
-      this happens, the preedit of IBusM17NEngine disappears for a
-      very short moment until it is shown again later when
-      IBusM17NEngine continues processing the key event. This may
-      cause a flicker in the preedit.  How visible this flicker is
-      depends on the circumstances, but sometimes it can be quite
-      visible and annoying. It seems possible to avoid 99% of the
-      flicker by calling ibus_engine_show_preedit_text ((IBusEngine
-      *)m17n) directly before **and** after
-      IBUS_ENGINE_CLASS(parent_class)->process_key_event() if the
-      preedit of IBusM17NEngine is not empty.
+      calls ibus_engine_simple_process_key_event(). This will handle compose sequences.
     */
-    if (mtext_len (m17n->context->preedit) > 0)
-        ibus_engine_show_preedit_text ((IBusEngine *)m17n);
     if (IBUS_ENGINE_CLASS (parent_class)->process_key_event (engine, keyval, keycode, modifiers)) {
         if (mtext_len (m17n->context->preedit) > 0) {
             gchar *buf;
@@ -783,18 +858,10 @@ ibus_m17n_engine_process_key_event (IBusEngine     *engine,
         }
         return TRUE;
     }
-    if (mtext_len (m17n->context->preedit) > 0)
-        ibus_engine_show_preedit_text ((IBusEngine *)m17n);
 
     if (modifiers & IBUS_RELEASE_MASK)
         return FALSE;
 
-    if (klass->use_us_layout) {
-        keyval = ibus_keymap_lookup_keysym (m17n->us_keymap,
-                                            keycode,
-                                            modifiers);
-    }
-
     MSymbol m17n_key = ibus_m17n_key_event_to_symbol (m17n,
                                                       keycode,
                                                       keyval,
